/************************************************************
 **
 ** pf_cmd.c : p-code cmd list functions
 **
 ** Copyright (c) 2005, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#include <assert.h>
#include <string.h>

#include "cmd_op.h"
#include "pfiled.h"
#include "pf_proc.h"
#include "pf_cmd.h"

/*
 * NAME
 *   pfile_cmd_add
 *
 * DESCRIPTION
 *   add a command to the command list
 *
 * PARAMETERS
 *
 * RETURN
 *
 * NOTES
 */
void pfile_cmd_add(pfile_t *pf, cmd_t cmd)
{
#if 1
  if (pfile_codegen_disable_get(pf)) {
    cmd_free(cmd);
  } else {
#else
    {
#endif
    pfile_pos_t    pos;
    pfile_block_t *blk;

    pos = pf->statement_start;
    if (pos.src) {
      cmd_line_set(cmd, pos.line);
      cmd_source_set(cmd, pos.src);
    }
    if (pf->cmd_cursor) {
      /* this cmd goes *after* pf->cmd_cursor */
      cmd_link_set(cmd, cmd_link_get(pf->cmd_cursor));
      cmd_link_set(pf->cmd_cursor, cmd);
      pf->cmd_cursor = cmd;
    } else {
      cmd_link_set(pf->cmd_tail, cmd);
      pf->cmd_tail = cmd;
    }
    if (!pf->cmd_head) {
      pf->cmd_head = cmd;
    }
    blk = pfile_proc_block_active_get(pfile_proc_active_get(pf));
    if (blk && (CMD_NONE == pfile_block_cmd_get(blk))) {
      assert(cmd_type_block_start == cmd_type_get(cmd));
      pfile_block_cmd_set(blk, cmd);
    }
  }
}

void pfile_cmd_special_add(pfile_t *pf, cmd_type_t type, pfile_proc_t *proc)
{
  cmd_t cmd;

  cmd = cmd_alloc(type, 0);
  if (cmd) {
    if (proc) {
      cmd_proc_set(cmd, proc);
    }
    pfile_cmd_add(pf, cmd);
  }
}

/*
 * NAME
 *   pfile_cmd_branch_add
 *
 * DESCRIPTION
 *   add a branching command
 *
 * PARAMETERS
 *   pf : pfile
 *   label : line label
 *   type  : branching type
 *   cond  : branching condition
 *   dst   : destination label
 *   var   : condition variable
 *
 * RETURN
 *
 * NOTES
 */
void pfile_cmd_branch_add(pfile_t *pf, 
  cmd_branchtype_t type, cmd_branchcond_t cond, label_t dst,
  value_t var, value_t proc, value_t *proc_params)
{
  cmd_t ptr;

  if ((cmd_branchcond_none != cond) 
    && value_is_const(var)) {
    /* this is an absolute branch disguised as a conditonal one */
    variable_const_t n;

    n = value_const_get(var);
    if (((0 == n) && (cmd_branchcond_false == cond)) 
        || ((0 != n) && (cmd_branchcond_true == cond))) {
      cond = cmd_branchcond_none;
    } else {
      type = cmd_branchtype_none;
    }
  }
  if (cmd_branchtype_none == type) {
    /* do nothing */
  } else {
    if (proc) {
      if (variable_def_type_function == value_type_get(proc)) {
        /* add this to the called list */
        pfile_proc_t *pproc;

        pproc = value_proc_get(proc);
        if (pproc) {
          pfile_proc_flag_set(value_proc_get(proc), PFILE_PROC_FLAG_DIRECT);
          pfile_proc_call_add(pfile_proc_active_get(pf), value_proc_get(proc));
        }
      }
    }
    ptr = cmd_branch_alloc(type, cond, dst, var, proc, proc_params);
    if (!ptr) {
      pfile_log_syserr(pf, result_memory);
    } else {
      pfile_cmd_add(pf, ptr);
    }
  }
}

/*
 * NAME
 *   pfile_cmd_label_add
 *
 * DESCRIPTION
 *   add a label only
 *
 * PARAMETERS
 *   pf    :
 *   label :
 *
 * RETURN
 *   none
 *
 * NOTES
 *   this doesn't really do anything other than set the
 *   ``next command gets this label''
 */
void pfile_cmd_label_add(pfile_t *pf, label_t label)
{
  cmd_t cmd;

  cmd = cmd_label_alloc(label);
  pfile_cmd_add(pf, cmd);
}

/*
 * NAME
 *   pfile_cmd_remove_unreachable
 *
 * DESCRIPTION
 *   remove unreachable commands
 *
 * PARAMETERS
 *   pf : pfile
 *
 * RETURN
 *   none
 *
 * NOTES
 *   the commands are removed because doing so reduces the assign/use
 *   counts. a later pass is made to remove any assignments to
 *   variables that are not used, and to change any uses of unassigned
 *   variables to 0
 */
void pfile_cmd_remove_unreachable(pfile_t *pf)
{
  cmd_t cmd;
  cmd_t cmd_pv;
  unsigned ct;
  boolean_t was_last_eos;

  cmd = pf->cmd_head;
  cmd_pv = 0;
  was_last_eos = boolean_true;

  /* first free unused commands */
  ct = 0;
  while (cmd) {
    label_t lbl;

    lbl = cmd_label_get(cmd);
    if (!cmd_is_reachable(cmd)
      || (lbl && !label_usage_get(lbl))
      || (was_last_eos && (cmd_type_statement_end == cmd_type_get(cmd)))) {
      if (cmd_pv) {
        cmd_link_set(cmd_pv, cmd_link_get(cmd));
      } else {
        pf->cmd_head = cmd_link_get(cmd);
        if (!pf->cmd_head) {
          pf->cmd_tail = 0;
        }
      }
      cmd_free(cmd);
      cmd = cmd_pv;
      ct++;
      was_last_eos = boolean_true;
    } else if (cmd_type_label != cmd_type_get(cmd)) {
      was_last_eos = cmd_type_statement_end == cmd_type_get(cmd);
    }
    cmd_pv = cmd;
    cmd = cmd_link_get(cmd);
  }
  pfile_log(pf, pfile_log_debug, "%u unreachable commands freed", ct);
}


/*
 * NAME
 *   pfile_cmd_remove_assignments
 *
 * DESCRIPTION
 *   remove all assignments to variables that are never used
 *
 * PARAMETERS
 *   pf : pfile handle
 *
 * RETURN
 *   none
 *
 * NOTES
 *   this is iterative -- at each iteration it removes all variables that are
 *   assigned but never used. it repeats until no more statements have been
 *   removed
 *   nb : a special case is made for assignment to a value with no
 *        corresponding variable. this should *only* happen when setting
 *        the return value for a function!
 */
void pfile_cmd_remove_assignments(pfile_t *pf)
{
  boolean_t changed;
  unsigned  ct;
  unsigned  pass;

  ct   = 0;
  pass = 0;
  do {
    cmd_t cmd;
    cmd_t cmd_pv;

    changed = boolean_false;
    pass++;

    cmd = pf->cmd_head;
    cmd_pv = 0;

    /* first free unused commands */
    while (cmd) {
      if (cmd_type_operator == cmd_type_get(cmd)) {
        value_t    dst;
        variable_t vdst;
        variable_t mvdst;

        dst   = cmd_opdst_get(cmd);
        vdst  = value_variable_get(dst);
        mvdst = variable_master_get(vdst);

        if (dst
          && vdst
          && !value_is_volatile(dst)
          && !value_is_used(dst)
          && !variable_is_used(mvdst)) {
          /* this is an assignment to an unused variable, remove it! */
          if (cmd_pv) {
            cmd_link_set(cmd_pv, cmd_link_get(cmd));
          } else {
            pf->cmd_head = cmd_link_get(cmd);
            if (!pf->cmd_head) {
              pf->cmd_tail = 0;
            }
          }
          cmd_free(cmd);
          cmd = cmd_pv;
          ct++;
          changed = boolean_true;
        }
      }
      cmd_pv = cmd;
      cmd = cmd_link_get(cmd);
    }
  } while (changed);
  pfile_log(pf, pfile_log_debug, "%u assignment commands freed in %u passes", ct, pass);
}

/* remove all sequences of {block} [{eos}] {end-of-block}
 * note: this must be repeated to catch nested blocks */

void pfile_cmd_remove_empty_blocks(pfile_t *pf)
{
  boolean_t changed;
  unsigned  ct;
  unsigned  pass;

  ct   = 0;
  pass = 0;
  do {
    cmd_t cmd;
    cmd_t cmd_pv;

    changed = boolean_false;
    pass++;

    cmd = pf->cmd_head;
    cmd_pv = 0;

    /* first free unused commands */
    while (cmd) {
      if (cmd_type_block_start == cmd_type_get(cmd)) {
        size_t remove_ct;

        remove_ct = 0;
        if ((cmd_type_statement_end == cmd_type_get(cmd_link_get(cmd)))
          && (cmd_type_block_end == cmd_type_get(cmd_link_get(
                cmd_link_get(cmd))))) {
          remove_ct = 3;
        } else if (cmd_type_block_end == cmd_type_get(cmd_link_get(cmd))) {
          /* {block} {block end} */
          remove_ct = 2;
        }
        if (remove_ct) {
          cmd_t cmd_ptr;

          ct++;
          changed = boolean_true;
          cmd_ptr = CMD_NONE;
          while (remove_ct--) {
            cmd_ptr = cmd_link_get(cmd);
            cmd_free(cmd);
            cmd = cmd_ptr;
          }
          cmd_link_set(cmd_pv, cmd_ptr);
          cmd = cmd_pv;
        }
      }
      cmd_pv = cmd;
      cmd = cmd_link_get(cmd);
    }
  } while (changed);
  pfile_log(pf, pfile_log_debug, "%u empty blocks freed in %u passes", ct, 
      pass);
}


void pfile_cmdlist_check(cmd_t cmd)
{
#if 0
  while (cmd) {
    pfile_proc_t *proc;

    assert(1368 != cmd_brval_get(cmd));
    assert(1368 != cmd_opdst_get(cmd));
    assert(1368 != cmd_opval1_get(cmd));
    assert(1368 != cmd_opval2_get(cmd));
    proc = cmd_brproc_get(cmd);
    if (proc) {
      unsigned ii;

      for (ii = 0; ii < pfile_proc_param_ct_get(proc); ii++) {
        assert(1368 != cmd_brproc_params_get(cmd)[ii]);
      }
    }
    cmd = cmd_link_get(cmd);
  }
#endif
}

